'use strict';

var meriyah = require('meriyah');
var recast = require('recast');

const { namedTypes: namedTypes$1, builders: builders$1 } = recast.types;
const {
  ArrayExpression, CallExpression, ExpressionStatement, FunctionExpression,
  Identifier, Literal, ObjectExpression
} = namedTypes$1;
const { arrayExpression } = builders$1;

class ConvertError extends Error {
  constructor(message) {
    super(message);
    this.name = 'ConvertError';
  }
}

function checkDeps(deps) {
  for (const dep of deps.elements) {
    if (!Literal.check(dep)) throw new ConvertError('dependency not a literal')
    if (typeof dep.value !== 'string') throw new ConvertError('dependency not a string')
  }
}

function checkParams(func) {
  for (const param of func.params) {
    if (!Identifier.check(param)) throw new ConvertError('argument not an identifier')
  }
}

function analyseStatement(statement) {
  if (!ExpressionStatement.check(statement)) throw new ConvertError('no expression statement')
  const { expression } = statement;
  if (!CallExpression.check(expression)) throw new ConvertError('no call expression')

  const { arguments: args } = expression;
  const { length } = args;
  if (length === 0) throw new ConvertError('no call arguments')

  const { callee } = expression;
  if (!Identifier.check(callee)) throw new ConvertError('callee not an identifier')

  // define('name', [deps], factory)
  if (callee.name === 'define') {
    let deps = args[0];
    let factory;
    let name;
    if (ArrayExpression.check(deps)) {
      if (length === 1) throw new ConvertError('too few arguments')
      if (length > 2) throw new ConvertError('too many arguments')
      checkDeps(deps);
      factory = args[1];
    } else if (Literal.check(deps)) {
      if (length === 1) throw new ConvertError('too few arguments')
      if (length > 3) throw new ConvertError('too many arguments')
      name = deps.value;
      deps = args[1];
      if (ArrayExpression.check(deps)) {
        if (length === 2) throw new ConvertError('too few arguments')
        checkDeps(deps);
        factory = args[2];
      } else {
        if (length > 2) throw new ConvertError('too many arguments')
        factory = deps;
        deps = arrayExpression([]);
      }
    } else {
      if (length > 1) throw new ConvertError('too many arguments')
      factory = deps;
      deps = arrayExpression([]);
    }

    if (ObjectExpression.check(factory)) return { statement, deps, object: factory, name }

    if (!FunctionExpression.check(factory)) throw new ConvertError('factory neither function nor object')

    checkParams(factory);
    return { statement, deps, factory, name }
  }

  // require([deps], success, error)
  if (callee.name === 'require') {
    const [deps, callback, failure] = args;
    if (!ArrayExpression.check(deps)) throw new ConvertError('missing dependencies')
    if (length === 1) throw new ConvertError('missing callback')
    if (!FunctionExpression.check(callback)) throw new ConvertError('callback not a function')
    if (length > 3) throw new ConvertError('too many arguments')
    let warnings;
    if (length === 3) {
      if (!FunctionExpression.check(failure)) throw new ConvertError('error handler not a function')
      warnings = ['error handler ignored'];
    }

    checkDeps(deps);
    checkParams(callback);

    return { statement, deps, callback, warnings }
  }
}

// define with factory: { statement, deps, factory }
// define with object: { statement, deps, object }
// require: { statement, deps, callback, warnings }
function analyseAmd(program) {
  const { body } = program;
  const { length } = body;
  if (length === 0) throw new ConvertError('no statement')
  if (length !== 1) throw new ConvertError('multiple statements')

  return analyseStatement(body[0])
}

const { namedTypes, builders } = recast.types;
const { ReturnStatement } = namedTypes;
const {
  importDeclaration, importDefaultSpecifier, identifier, literal,
  exportDefaultDeclaration
} = builders;

function cloneString(source) {
  const { value } = source;
  const clone = literal(value);
  clone.extra = { raw: source.raw, rawValue: value };
  return clone
}

function convertToEsm(program, { statement, deps, factory, object, callback }) {
  const { params = [] } = factory || {};
  const imps = deps.elements.map((dep, index) => {
    const param = params[index];
    const spec = param ? [importDefaultSpecifier(identifier(param.name))] : [];
    const imp = importDeclaration(spec, cloneString(dep));
    imp.comments = dep.comments;
    return imp
  });

  let stats = [];
  let exps = [];
  if (factory) {
    stats = factory.body.body.filter(stat => {
      if (!ReturnStatement.check(stat)) return true
      const exp = exportDefaultDeclaration(stat.argument);
      exp.comments = stat.comments;
      exps = [exp];
    });
  } else if (object) {
    const exp = exportDefaultDeclaration(object);
    exp.comments = object.comments;
    exps = [exp];
  } else {
    stats = callback.body.body;
  }

  const { comments } = statement;
  if (comments) program.comments = [...(program.comments || []), ...comments];
  program.body = [...imps, ...stats, ...exps];
}

function ensureLineBreak (output, input) {
  if (input.endsWith('\n') && !output.endsWith('\n')) {
    output += '\n';
  }
  return output
}

function improveCode(output, input, warnings = []) {
  const lines = output.split(/\r?\n/);
  let imported;
  let separated;
  let emptied;
  for (let i = 0, l = lines.length; i < l; ++i) {
    const line = lines[i];
    if (!emptied && line === ';') {
      emptied = true;
      lines.splice(i--, 1);
      --l;
      warnings.push('"use strict" removed');
    } else if (!separated) {
      if (line.startsWith('import')) {
        imported = true;
      } else if (imported) {
        separated = true;
        if (line.length) {
          lines.splice(i++, 0, '');
          ++l;
          warnings.push('imports separated');
        }
      }
    }
  }
  const code = ensureLineBreak(lines.join('\n'), input);
  return { code, warnings }
}

const convertComment = comment => {
  let { type } = comment;
  type = type[0] === 'S' ? 'Line' : 'Block';
  return { ...comment, type }
};

const parser = {
  parse(source) {
    const comments = [];
    const tokens = [];
    const ast = meriyah.parse(source, {
      module: true, next: true, raw: true, loc: true,
      onComment: comments, onToken: tokens
    });
    ast.comments = comments.map(convertComment);
    ast.tokens = tokens;
    return ast
  }
};

function convert(source) {
  const ast = recast.parse(source, { parser });
  const { program } = ast;

  const analysis = analyseAmd(program);
  const { warnings = [] } = analysis;

  convertToEsm(program, analysis);
  let { code } = recast.print(ast);
  ({ code } = improveCode(code, source, warnings));
  return { code, warnings, name: analysis.name }
}

exports.convert = convert;
//# sourceMappingURL=index.cjs.map
